package com.sillycycle.bagleyd.tetris;

/*
 * @(#)PolyominoThing.java
 *
 * Copyright 2016 - 2020  David A. Bagley, bagleyd AT verizon.net
 *
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and
 * that both that copyright notice and this permission notice appear in
 * supporting documentation, and that the name of the author not be
 * used in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.
 *
 * This program is distributed in the hope that it will be "useful",
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */

import java.io.File;
import java.io.IOException;
import java.net.URL;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.ParserConfigurationException;
import org.w3c.dom.DOMException;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;
import org.w3c.dom.Node;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;

/**
 * The <code>PolyominoThing</code> class handles the objects to be
 * manipulated for the <code>TetrisdFrame</code> class.
 *
 * @author	David A. Bagley
 * @author	bagleyd AT verizon.net
 * @author	https://www.sillycycle.com/altris.html
 */

public class PolyominoThing {
	static final int NODIAG = 0;   /* Normal, ie no diagonals */
	static final int DIAGONAL = 1; /* Normal and diagonal */
	static final int NOMIX = 0;    /* Do not mix lowered number units */
	static final int MIXED = 1;
	static final int MAX_MODES = 2;
	static final int NONE = 0;   /* No rotation or reflection */
	static final int NOREFL = 1; /* Normal, no reflection allow rotation */
	static final int ALL = 2;    /* Allow reflection or rotation */
	static final int MAX_TURN_TYPES = 3;
	static final int JUMPIN = 0;
	static final int GRADUAL = 1;
	static final int MAX_START_TYPES = 2;
	static final int DEFAULT_UNITS = 4;
	static final int MIN_UNITS = 2; /* Why not 1 you say? */
	static final int MAX_UNITS = 6;
	static final int DEFAULT_DIAGONALUNITS = 3;
	static final int MIN_DIAGONALUNITS = MIN_UNITS;
	static final int MAX_DIAGONALUNITS = 3;
	static final int MAX_POLYOMINOES = 232; /* 216 + 16 DIAGONAL */
	static final int MAX_START_POLYOMINOES = 59; /* 55 + 4 DIAGONAL */

	static final boolean DEBUG = false;
	static int toss = 0;

	static PolyMode[] tris;
	static int[] counter;
	static int[][][] startCounter;

	public PolyominoThing() {
		counter = new int[PolyominoThing.MAX_UNITS -
			PolyominoThing.MIN_UNITS + 1];
		startCounter = new int[PolyominoThing.MAX_UNITS -
			PolyominoThing.MIN_UNITS + 1]
			[PolyominoThing.MAX_MODES][PolyominoThing.MAX_TURN_TYPES];
		tris = new PolyMode[PolyominoThing.MAX_UNITS -
			PolyominoThing.MIN_UNITS + 1];
		for (int pi = 0; pi < (PolyominoThing.MAX_UNITS -
				PolyominoThing.MIN_UNITS + 1); pi++)
			tris[pi] = new PolyMode(PolyominoThing.MAX_MODES,
				PolyominoThing.MAX_TURN_TYPES,
				PolyominoThing.MAX_START_POLYOMINOES,
				PolyominoThing.MAX_POLYOMINOES,
				PolyominoThing.MAX_UNITS, PolyominoThing.MAX_UNITS);
		for (int pi = 0; pi <= PolyominoThing.MAX_UNITS -
		     PolyominoThing.MIN_UNITS; pi++) {
		  counter[pi] = 0;
		  for (int polyomino = 0; polyomino <
		       PolyominoThing.MAX_POLYOMINOES; polyomino++)
		    for (int j = 0; j < PolyominoThing.MAX_UNITS; j++)
		      for (int i = 0; i < PolyominoThing.MAX_UNITS; i++)
			tris[pi].poly[polyomino].shape[j][i] = 0;
		  for (int j = 0; j < PolyominoThing.MAX_TURN_TYPES; j++) {
		    for (int polyomino = 0; polyomino <
			 PolyominoThing.MAX_START_POLYOMINOES; polyomino++) {
		      tris[pi].mode[PolyominoThing.NODIAG].start[polyomino][j] = 0;
		      tris[pi].mode[PolyominoThing.DIAGONAL].start[polyomino][j] = 0;
		    }
		    startCounter[pi][PolyominoThing.NODIAG][j] = 0;
		    startCounter[pi][PolyominoThing.DIAGONAL][j] = 0;
		    tris[pi].mode[PolyominoThing.NODIAG].number[PolyominoThing.NOMIX][j] = 0;
		    tris[pi].mode[PolyominoThing.DIAGONAL].number[PolyominoThing.NOMIX][j] = 0;
		    tris[pi].mode[PolyominoThing.NODIAG].number[PolyominoThing.MIXED][j] = 0;
		    tris[pi].mode[PolyominoThing.DIAGONAL].number[PolyominoThing.MIXED][j] = 0;
		  }
		  tris[pi].mode[PolyominoThing.NODIAG].turnStyle = PolyominoThing.NONE;
		  tris[pi].mode[PolyominoThing.DIAGONAL].turnStyle = PolyominoThing.NONE;
		}
	}

	static PolyMode[] readObjectsXML(String fileName, URL url) {
		try {
			DocumentBuilderFactory docBuilderFactory =
				DocumentBuilderFactory.newInstance();
			DocumentBuilder docBuilder =
				docBuilderFactory.newDocumentBuilder();
			Document doc;

			if (url == null) {
				doc = docBuilder.parse(new File(fileName));
			} else {
				doc = docBuilder.parse(new URL(url, fileName).openStream());
				//System.out.println("Loaded xml: " + url + fileName);
			}
			// normalize text representation
			doc.getDocumentElement().normalize();
			if (DEBUG)
				System.out.println("Root element of the doc is " +
					doc.getDocumentElement().getNodeName());
			NodeList header =
				doc.getElementsByTagName("header");
			readHeader(header.item(0));
			NodeList objectsList =
				doc.getElementsByTagName("objects");
			for (int o = 0; o < objectsList.getLength(); o++) {
				readObjects(objectsList.item(o));
			}
			return tris;
		} catch (SAXParseException err) {
			System.out.println("** Parsing error" + ", line " +
				err.getLineNumber() + ", uri " +
				err.getSystemId());
			System.out.println(" " + err.getMessage());
		} catch (SAXException e) {
			Exception x = e.getException();
			((x == null) ? e : x).printStackTrace();
		} catch (ParserConfigurationException | IOException | DOMException | NumberFormatException e) {
			e.printStackTrace();
		}
		return null;
	}

	static void readHeader(Node headerNode) {
		try {
			if (headerNode.getNodeType() != Node.ELEMENT_NODE) {
				return;
			}
			Element polyTypeElement = (Element) headerNode;
			NodeList polyTypeList = polyTypeElement.getElementsByTagName("polyType");
			for (int pi = 0; pi < polyTypeList.getLength(); pi++) {
				int[] polyType = readPolyType(polyTypeList.item(pi));
				int index = 0;
				if (polyType == null)
					return;
				int size = polyType[index++];
				if (size != pi + MIN_UNITS) {
					System.out.println("index: " + (pi + MIN_UNITS) + " != size: " + size);
					return;
				}
				for (int i = 0; i < MAX_TURN_TYPES; i++) {
					tris[pi].mode[NODIAG].number[NOMIX][i] = polyType[index++];
				}
				tris[pi].mode[NODIAG].turnStyle = polyType[index++];
				if (polyType.length > index) {
					tris[pi].diagonalSwitch = 1;
					for (int i = 0; i < MAX_TURN_TYPES; i++) {
						tris[pi].mode[DIAGONAL].number[NOMIX][i] = polyType[index++];
					}
					tris[pi].mode[DIAGONAL].turnStyle = polyType[index++];
				} else {
					tris[pi].diagonalSwitch = 0;
				}
			}
		} catch (DOMException | NumberFormatException e) {
			e.printStackTrace();
		}
	}

	static int[] readPolyType(Node polyTypeNode) {
		try {
			if (polyTypeNode.getNodeType() != Node.ELEMENT_NODE) {
				return null;
			}
			Element polyTypeElement = (Element) polyTypeNode;
			NodeList itemList = polyTypeElement.getElementsByTagName("size");
			Element itemElement = (Element) itemList.item(0);
			NodeList textItemList = itemElement.getChildNodes();
			int size = Integer.parseInt((textItemList.item(0)).getNodeValue().trim());
			if (DEBUG)
				System.out.println(" size: " + size);

			itemList = polyTypeElement.getElementsByTagName("kinds");
			if (itemList.getLength() > 1) {
				System.out.println(" kinds = " + itemList.getLength());
			}
			int[] kindsArray = readKinds(itemList.item(0));
			int kindsCount = 0;
			if (kindsArray != null) {
				if (DEBUG)
					System.out.println("kinds: " + kindsArray.length);
				kindsCount = kindsArray.length;
			}

			itemList = polyTypeElement.getElementsByTagName("diagonalKinds");
			if (itemList.getLength() > 1) {
				System.out.println(" diagonalKinds = " + itemList.getLength());
			}
			int[] diagonalKindsArray = readKinds(itemList.item(0));
			int diagonalKindsCount = 0;
			if (diagonalKindsArray != null) {
				if (DEBUG)
					System.out.println("diagonalKinds: " + diagonalKindsArray.length);
				diagonalKindsCount = kindsCount;
			}

			int[] total = new int[1 + kindsCount + diagonalKindsCount];
			total[0] = size;
			System.arraycopy(kindsArray, 0, total, 1, kindsCount);
			if (diagonalKindsArray != null) {
				for (int i = 0; i < diagonalKindsCount; i++) {
					total[i + 1 + kindsCount] = diagonalKindsArray[i];
				}
			}
			return total;
		} catch (DOMException | NumberFormatException e) {
			e.printStackTrace();
			return null;
		}
	}

	static int[] readKinds(Node kindsNode) {
		try {
			if (kindsNode.getNodeType() != Node.ELEMENT_NODE) {
				return null;
			}
			int[] kind = new int[MAX_TURN_TYPES + 1];
			Element kindsElement = (Element) kindsNode;
			NodeList itemList;
			Element itemElement;
			NodeList textItemList;
			itemList = kindsElement.getElementsByTagName("distinctNoRotations");
			itemElement = (Element) itemList.item(0);
			if (itemElement == null)
				return null;
			textItemList = itemElement.getChildNodes();
			kind[0] = Integer.parseInt((textItemList.item(0)).getNodeValue().trim());
			itemList = kindsElement.getElementsByTagName("distinctNoReflections");
			itemElement = (Element) itemList.item(0);
			if (itemElement == null)
				return null;
			textItemList = itemElement.getChildNodes();
			kind[1] = Integer.parseInt((textItemList.item(0)).getNodeValue().trim());
			itemList = kindsElement.getElementsByTagName("distinct");
			itemElement = (Element) itemList.item(0);
			if (itemElement == null)
				return null;
			textItemList = itemElement.getChildNodes();
			kind[2] = Integer.parseInt((textItemList.item(0)).getNodeValue().trim());
			itemList = kindsElement.getElementsByTagName("style");
			itemElement = (Element) itemList.item(0);
			if (itemElement == null)
				return null;
			textItemList = itemElement.getChildNodes();
			kind[3] = Integer.parseInt((textItemList.item(0)).getNodeValue().trim());
			if (DEBUG)
				System.out.println("style: " + kind[3]);
			return kind;
		} catch (DOMException | NumberFormatException e) {
			e.printStackTrace();
			return null;
		}
	}

	static void readObjects(Node objectsNode) {
		try {
			if (objectsNode.getNodeType() != Node.ELEMENT_NODE) {
				return;
			}
			Element geometryElement = (Element) objectsNode;
			NodeList geometryList = geometryElement.getElementsByTagName("geometry");
			for (int g = 0; g < geometryList.getLength(); g++) {
				readGeometry(geometryList.item(g));
			}
			for (int i = 0; i <= PolyominoThing.MAX_UNITS -
					PolyominoThing.MIN_UNITS; i++)
				for (int j = 0; j < PolyominoThing.MAX_TURN_TYPES; j++)
					for (int k = 0; k <= i; k++) {
						tris[i].mode[PolyominoThing.NODIAG].number
							[PolyominoThing.MIXED][j] +=
						tris[k].mode[PolyominoThing.NODIAG].number
							[PolyominoThing.NOMIX][j];
						if (tris[i].diagonalSwitch != 0)
						/*since k & i are not independent*/
						tris[i].mode[PolyominoThing.DIAGONAL].number
							[PolyominoThing.MIXED][j] +=
						tris[k].mode[PolyominoThing.DIAGONAL].number
							[PolyominoThing.NOMIX][j];
					}
		} catch (DOMException | NumberFormatException e) {
			e.printStackTrace();
		}
	}

	static void readGeometry(Node geometryNode) {
		try {
			if (geometryNode.getNodeType() != Node.ELEMENT_NODE) {
				return;
			}
			Element itemElement = (Element) geometryNode;
			NodeList itemList = itemElement.getElementsByTagName("size");
			itemElement = (Element) itemList.item(0);
			if (itemElement == null)
				return;
			NodeList textItemList = itemElement.getChildNodes();
			int size = Integer.parseInt((textItemList.item(0)).getNodeValue().trim());
			if (DEBUG)
				System.out.println("geo size: " + size);
			Element polyGeometryElement = (Element) geometryNode;
			NodeList polyGeometryList = polyGeometryElement.getElementsByTagName("polyGeometry");
			for (int pg = 0; pg < polyGeometryList.getLength(); pg++) {
				readPolyGeometry(size, polyGeometryList.item(pg));
			}
		} catch (DOMException | NumberFormatException e) {
			e.printStackTrace();
		}
	}

	static void readPolyGeometry(int size, Node polyGeometryNode) {
		try {
			if (polyGeometryNode.getNodeType() != Node.ELEMENT_NODE) {
				return;
			}
			int polyKindNumber = 0;
			Element polyGeometryElement = (Element) polyGeometryNode;
			NodeList itemList;
			Element itemElement;
			NodeList textItemList;
			itemList = polyGeometryElement.getElementsByTagName("polyNumber");
			itemElement = (Element) itemList.item(0);
			textItemList = itemElement.getChildNodes();
			int pi = Integer.parseInt((textItemList.item(0)).getNodeValue().trim()) -
				PolyominoThing.MIN_UNITS;

			itemList = polyGeometryElement.getElementsByTagName("polyKindNumber");
			itemElement = (Element) itemList.item(0);
			textItemList = itemElement.getChildNodes();
			if (textItemList.item(0) != null) {
				polyKindNumber = Integer.parseInt((textItemList.item(0)).getNodeValue().trim());
			}
			itemList = polyGeometryElement.getElementsByTagName("diagonalPolyKindNumber");
			itemElement = (Element) itemList.item(0);
			textItemList = itemElement.getChildNodes();
			if (textItemList.item(0) != null && tris[pi].diagonalSwitch != 0) {
				polyKindNumber = Integer.parseInt((textItemList.item(0)).getNodeValue().trim());
			}

			itemList = polyGeometryElement.getElementsByTagName("data");
			itemElement = (Element) itemList.item(0);
			if (itemElement == null)
				return;
			textItemList = itemElement.getChildNodes();
			String xmlInt = (textItemList.item(0)).getNodeValue();
			String newInt = xmlInt.replaceAll("\\r|\\n", " ").replaceAll("\\s+", " ").trim();
			//if (DEBUG)
			//	System.out.println("Data: " + newInt);
			String[] numbers = newInt.split(" ");
			int index = 0;
			for (int polyomino = 0; polyomino < polyKindNumber - toss; polyomino++) {
				//System.out.println(polyKindNumber);
				int sum = polyomino + counter[pi];
				index++;
				tris[pi].poly[sum].rotation =
					Integer.parseInt(numbers[index++])
					+ counter[pi] - toss;
				tris[pi].poly[sum].reflection =
					Integer.parseInt(numbers[index++])
					+ counter[pi] - toss;
				for (int initHeight = PolyominoThing.JUMPIN;
						initHeight <= PolyominoThing.GRADUAL; initHeight++) {
					tris[pi].poly[sum].startHeight[(initHeight == PolyominoThing.JUMPIN)
						? 1 : 0] = Integer.parseInt(numbers[index++]);
					int start = Integer.parseInt(numbers[index++]);
					for (int diag = PolyominoThing.NODIAG;
							diag <= tris[pi].diagonalSwitch; diag++) {
						if (initHeight == PolyominoThing.JUMPIN) {
							if (pi == 0 ||
									tris[pi - 1].mode[diag].turnStyle ==
									PolyominoThing.NONE) {
								int i = startCounter[pi][diag][PolyominoThing.NONE];
								tris[pi].mode[diag].start[i]
									[PolyominoThing.NONE] = sum;
								startCounter[pi][diag][PolyominoThing.NONE]++;
							}
							if ((pi == 0 ||
									tris[pi - 1].mode[diag].turnStyle <
									PolyominoThing.ALL) && start !=
									PolyominoThing.NONE) {
								int i = startCounter[pi][diag]
									[PolyominoThing.NOREFL];
								tris[pi].mode[diag].start[i]
									[PolyominoThing.NOREFL] = sum;
								startCounter[pi][diag][PolyominoThing.NOREFL]++;
							}
							if (start == PolyominoThing.ALL) {
								int i = startCounter[pi][diag][PolyominoThing.ALL];
								tris[pi].mode[diag].start[i]
									[PolyominoThing.ALL] = sum;
								startCounter[pi][diag][PolyominoThing.ALL]++;
							}
						}
					}
				}
				tris[pi].poly[sum].size = size;
				for (int j = 0; j < size; j++)
					for (int i = 0; i < size; i++) {
						tris[pi].poly[sum].shape[j][i] =
							Integer.parseInt(numbers[index++]);
					}
			}
			//if (DEBUG)
			//	System.out.println("counter " + counter[pi] + " " + polyKindNumber);
			counter[pi] += polyKindNumber - toss;
			toss = 0;
		} catch (DOMException | NumberFormatException e) {
			e.printStackTrace();
			//return null;
		}
	}
}
