package com.sillycycle.bagleyd.tetris;

/*
 * @(#)TetrisCanvas.java
 *
 * Copyright 1993 - 2020  David A. Bagley, bagleyd AT verizon.net
 *
 * Taken from GENERIC X-BASED TETRIS
 *
 * Copyright 1992 - 1995 Q. Alex Zhao, azhao AT cc.gatech.edu
 *
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and
 * that both that copyright notice and this permission notice appear in
 * supporting documentation, and that the name of the author not be
 * used in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.
 *
 * This program is distributed in the hope that it will be "useful",
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */

import java.awt.Canvas;
import java.awt.Color;
import java.awt.Graphics;
import java.awt.Point;
import java.util.Random;
import javax.swing.JFrame;

/**
 * The <code>TetrisCanvas</code> class does all the graphics for
 * <code>TetrisFrame</code> class.
 *
 * @author	David A. Bagley
 * @author	bagleyd AT verizon.net
 * @author	https://www.sillycycle.com/altris.html
 */

public class TetrisCanvas extends Canvas {
	private static final long serialVersionUID = 42L;

	transient PolyMode[]  tris; //[PolyominoThing.MAX_UNITS - PolyominoThing.MIN_UNITS + 1];

	Random generator;
	int startlevel = TetrisInterface.DEFAULT_LEVEL;
	int level = TetrisInterface.DEFAULT_LEVEL;
	int columns = TetrisInterface.DEFAULT_COLUMNS;
	int rows = TetrisInterface.DEFAULT_ROWS;
	int fill = TetrisInterface.DEFAULT_FILL;
	boolean fixedFill = false, suddenAppear = false;
	boolean bonus = false;
	boolean cw = true, showNext = false;
	boolean invert = false, dropPaused = false;
	boolean firstPaint = false, resizePaint = false;
	boolean redrawPaint = false, framePaint = false;
	boolean focus = true;
	int score = 0, completedRows = 0;
	int boxSize = TetrisInterface.DEFAULT_BOXSIZE;
	int boxWidth = boxSize, boxHeight = boxSize;
	int viewLength = boxWidth * columns;
	Point fieldOffset = new Point(2, 2);
	Color background;
	transient Thing curThing, nextThing;
	transient Spec curSpec = new Spec(PolyominoThing.DEFAULT_UNITS);
	transient Spec nextSpec = new Spec(PolyominoThing.DEFAULT_UNITS);
	PolyominoImage polyominoImage;
	transient Field[][] field = null;
	JFrame frame = null;
	boolean debug = false;

	public TetrisCanvas() {
		curThing = new Thing();
		nextThing = new Thing();
		polyominoImage = new PolyominoImage();
		generator = new Random(System.nanoTime());
		readPolyominoes();
		sizeInit();
		inits();
	}

	void specInit(int units, boolean diagonal, boolean mixed) {
		curSpec = new Spec(units, diagonal, mixed, false);
		nextSpec = new Spec(units, diagonal, mixed, false);
	}

	final void sizeInit() {
		field = new Field[rows][columns];
		for (int y = 0; y < rows; y++)
			for (int x = 0; x < columns; x++)
				field[y][x] = new Field();
	}

	final void inits() {
		int turnStyle, numberPolyominoes;

		turnStyle = tris[nextSpec.units - PolyominoThing.MIN_UNITS].mode
			[(nextSpec.diagonal) ? 1 : 0].turnStyle;
		if (turnStyle < 0 || turnStyle >= PolyominoThing.MAX_TURN_TYPES) {
			System.err.println("Tetris: corrupted polyomino file");
			System.out.println("\tunits = " + nextThing.unitsIndex +
				", diagonal = " + nextSpec.diagonal +
				", turnStyle =" + turnStyle);
		}
		numberPolyominoes = tris[nextSpec.units - PolyominoThing.MIN_UNITS].mode
			[(nextSpec.diagonal) ? 1 : 0].number[(nextSpec.mixed) ? 1 : 0][turnStyle];
		if (numberPolyominoes <= 0 ||
				numberPolyominoes > PolyominoThing.MAX_START_POLYOMINOES) {
			System.err.println("Tetris: corrupted polyomino file");
			System.err.println("\tnumberPolyominoes = " + numberPolyominoes);
			System.out.println("\tunits = " + nextThing.unitsIndex +
				", diagonal = " + nextSpec.diagonal +
				", mixed =" + nextSpec.mixed +
				", turnStyle = " + turnStyle);
		}
		initFill();
		/* new things */
		newThing();
		newThing();
	}

	void newThing() {
		curThing.clone(nextThing);
		curSpec.clone(nextSpec);
		nextThing.setRandomNumber(Math.abs(generator.nextInt()));
		nextThing.setRotation(generator.nextInt(TetrisInterface.MAX_SIDES));
		nextThing.setReflection(generator.nextInt(2));
		nextThing.setBonus(nextSpec.bonus);
		redoNext();
	}

	void drawTitle() {
		//TetrisApp.drawTitle
	}

	void drawStatus() {
		/*
		TetrisApp.drawScore("Score: " + score);
		TetrisApp.drawLevel("Level: " + level);
		TetrisApp.drawRows("Rows: " + complededRows);
		*/
	}

	void resizeField() {
		boxWidth = (getSize().width - 4) / columns;
		boxHeight = ((getSize().height - 4) / 2) / columns;
		if (boxWidth > boxHeight)
			boxWidth = boxHeight;
		else
			boxHeight = boxWidth;
		viewLength = boxWidth * columns;
		fieldOffset = new Point(
			(getSize().width  + 1 - viewLength) / 2,
			(getSize().height + 1 - 2 * viewLength) / 2);
	}

	void resetImage() {
		if (this.boxWidth > 2) {
			polyominoImage.createBitmaps(false, false,
				this.boxWidth, background);
			polyominoImage.createBitmaps(true, false,
				this.boxWidth, background);
		}
	}

	void clearUnit(Graphics g, int i, int j) {
		PolyominoImage.clearUnits(g, background,
			fieldOffset.x, fieldOffset.y,
			i, j, boxWidth, boxHeight);
	}

	void xorUnit(Graphics g, int i, int j, int k) {
		if (field[j][i].pmid >= 0) {
			Color color = PolyominoImage.colorThings[field[j][i].cid];
			if (k % 2 == 0) // Simulate Xor
				color = new Color(255 - color.getRed(),
					255 - color.getGreen(),
					255 - color.getBlue());
			polyominoImage.drawUnit(g, false, field[j][i].pmid,
				color, fieldOffset.x, fieldOffset.y,
				i, j, boxWidth, boxHeight);
		}
	}

	void clearField() {
		Graphics g = this.getGraphics();

		// This will flash
		PolyominoImage.clearUnits(g, background, 0, 0, 0, 0,
			getSize().width, getSize().height);
		resizeField();
		resetImage();
		for (int i = 0; i < columns; i++)
			for (int j = 0; j < rows; j++)
				clearUnit(g, i, j);
		drawFrame(g, true);
	}

	void draw3DFrame(boolean inside, int x, int y, int width, int height) {
		Graphics g = getGraphics();

		if (g == null)
			return;
		try {
			if (inside)
				g.setColor(background.darker());
			else
				g.setColor(background);
			g.fillRect(x, y, width, 1);
			g.fillRect(x, y + 1, 1, height - 1);
			g.fillRect(x + 1, y + 1, width - 2, 1);
			g.fillRect(x + 1, y + 2, 1, height - 3);
			if (inside)
				g.setColor(background.brighter().brighter());
			else
				g.setColor(background);
			g.fillRect(x + 1, y + height - 1, width - 1, 1);
			g.fillRect(x + width - 1, y + 1, 1, height - 2);
			g.fillRect(x + 2, y + height - 2, width - 3, 1);
			g.fillRect(x + width - 2, y + 2, 1, height - 4);
		} finally {
			g.dispose();
		}
	}

	void drawFrame(Graphics g, boolean show) {
		int offsetX = fieldOffset.x - 2;
		int offsetY = fieldOffset.y - 2;
		int lengthX = fieldOffset.x + viewLength + 2;
		int lengthY = fieldOffset.y + 2 * viewLength + 2;

		if (g == null)
			return;
		g.setColor(background.darker().darker());
		g.fillRect(0, 0, getSize().width, offsetY);
		g.fillRect(0, offsetY, offsetX, lengthY - offsetY);
		g.fillRect(0, lengthY, getSize().width, offsetY);
		g.fillRect(lengthX, offsetY,
			getSize().width - lengthX, lengthY - offsetY);
		draw3DFrame(show && focus, offsetX, offsetY,
			lengthX - offsetX, lengthY - offsetY);
	}

	void drawFrame(boolean show) {
		Graphics g = getGraphics();

		if (g == null)
			return;
		try {
			drawFrame(g, show);
		} finally {
			g.dispose();
		}
	}

	void drawField(int start, int finish) {
		Graphics g = this.getGraphics();
		int i, j;

		for (i = 0; i < columns; i++)
			for (j = start; j < finish; j++) {
				if (field[j][i].pmid >= 0)
					polyominoImage.drawUnit(g, false,
						field[j][i].pmid,
						field[j][i].cid,
						fieldOffset.x, fieldOffset.y,
						i, j, boxWidth, boxHeight);
				else
					clearUnit(g, i, j);
			}
	}

	void drawField() {
		drawField(0, rows);
	}

	void drawThing() {
		Graphics g = this.getGraphics();
		int i, j;

		for (i = 0; i < curThing.size; i++)
			for (j = 0; j < curThing.size; j++) {
				int xi = curThing.position.x + i;
				int yj = curThing.position.y + j;

				if (yj >= 0 && tris[curThing.unitsIndex].poly
				      [curThing.polyNumber].shape[j][i] != 0) {

				  polyominoImage.drawUnit(g, false,
				    PolyominoImage.polyPixmap(i, j,
				      tris[curThing.unitsIndex].poly[curThing.polyNumber].shape,
				      curThing.size),
				      curThing.colorNumber,
				      fieldOffset.x, fieldOffset.y,
				      xi, yj, boxWidth, boxHeight);
				  /* This uses precomputed values */
				  /* polyominoImage.drawUnit(g, false,
				    PolyominoImage.NUM_BITMAPS - tris[curThing.unitsIndex].poly
				      [curThing.polyNumber].shape[j][i] - 1,
				      curThing.colorNumber,
				      fieldOffset.x, fieldOffset.y,
				      xi, yj, boxWidth, boxHeight); */
				}
			}
	}

	void drawThingDiff(Thing oldThing) {
		Graphics g = this.getGraphics();
		int i, j;

		for (i = 0; i < curThing.size; i++)
			for (j = 0; j < curThing.size; j++) {
				int xi = curThing.position.x + i;
				int yj = curThing.position.y + j;

				if (yj >= 0 && (tris[curThing.unitsIndex].poly
				       [curThing.polyNumber].shape[j][i] != 0))
				  polyominoImage.drawUnit(g, false,
						PolyominoImage.polyPixmap(i, j,
				    tris[curThing.unitsIndex].poly[curThing.polyNumber].shape,
				    curThing.size),
				    curThing.colorNumber,
				    fieldOffset.x, fieldOffset.y,
				    xi, yj, boxWidth, boxHeight);
				/* This uses precomputed values */
				/*  polyominoImage.drawUnit(g, false,
				    PolyominoImage.NUM_BITMAPS - tris[curThing.unitsIndex].poly
				      [curThing.polyNumber].shape[j][i] - 1,
				      curThing.colorNumber,
				      fieldOffset.x, fieldOffset.y,
				      xi, yj, boxWidth, boxHeight); */
			}
		for (i = 0; i < curThing.size; i++)
			for (j = 0; j < curThing.size; j++) {
				int xi = oldThing.position.x + i;
				int yj = oldThing.position.y + j;
				int ox = xi - curThing.position.x;
				int oy = yj - curThing.position.y;

				if (yj >= 0 && (tris[oldThing.unitsIndex].poly
				     [oldThing.polyNumber].shape[j][i] != 0) &&
				    ((ox < 0) || (ox >= curThing.size) ||
				     (oy < 0) || (oy >= curThing.size) ||
				     tris[curThing.unitsIndex].poly
				      [curThing.polyNumber].shape[oy][ox] == 0))
					clearUnit(g, xi, yj);
			}
	}

	void drawNext(JFrame app) {
		((TetrisFrame) app).drawNext(nextThing, tris);
	}

	void clearNext(JFrame app) {
		((TetrisFrame) app).clearNext(nextThing, tris);
	}

	void redoNext() {
		int diagonal = (nextSpec.diagonal) ? 1 : 0;
		int turnStyle =
			tris[nextSpec.units - PolyominoThing.MIN_UNITS].mode
				[diagonal].turnStyle;
		int nextStart, i, diffPolys;

		nextThing.unitsIndex = nextSpec.units - PolyominoThing.MIN_UNITS +
			((nextThing.getBonus()) ? 1 : 0);
		diffPolys = tris[nextThing.unitsIndex].mode[diagonal].number
			[(nextThing.getBonus()) ? PolyominoThing.NOMIX :
			((nextSpec.mixed) ? 1 : 0)][turnStyle];
		if (diffPolys == 0) {
			System.out.println("no poly units=" + nextThing.unitsIndex + PolyominoThing.MIN_UNITS
				+ ", diagonal=" + nextSpec.diagonal
				+ ", mixed=" + nextSpec.mixed
				+ ", bonus=" + nextThing.getBonus());
			return;
		}
		nextStart = nextThing.getRandomNumber() % diffPolys;
		nextThing.colorNumber = nextStart;
		if (nextSpec.mixed && !nextThing.getBonus()) {
			nextThing.unitsIndex = 0;
			while (nextStart >= tris[nextThing.unitsIndex].mode
				  [diagonal].number[PolyominoThing.NOMIX][turnStyle]) {
				nextStart -= tris[nextThing.unitsIndex].mode
				  [diagonal].number[PolyominoThing.NOMIX][turnStyle];
				nextThing.unitsIndex++;
			}
		}
		nextThing.polyNumber = tris[nextThing.unitsIndex].mode
			[diagonal].start[nextStart][turnStyle];
		if (tris[nextThing.unitsIndex].mode[diagonal].turnStyle >
				PolyominoThing.NONE)
			for (i = 0; i < nextThing.getRotation(); i++)
				nextThing.polyNumber =
					tris[nextThing.unitsIndex].poly
					[nextThing.polyNumber].rotation;
		if (tris[nextThing.unitsIndex].mode[diagonal].turnStyle ==
				PolyominoThing.ALL)
			for (i = 0; i < nextThing.getReflection(); i++)
				nextThing.polyNumber =
					tris[nextThing.unitsIndex].poly
					[nextThing.polyNumber].reflection;
		nextThing.size = tris[nextThing.unitsIndex].poly
			[nextThing.polyNumber].size;
		nextThing.position.x = generator.nextInt(columns - nextThing.size + 1);
		/* nextThing.position.x = (columns - nextThing.size) / 2; */
		nextThing.position.y = -tris[nextThing.unitsIndex].poly
			[nextThing.polyNumber].startHeight[(suddenAppear) ? 1 : 0];
		if (debug)
			System.out.println("Size: " + nextThing.size +
				", Poly: " + nextThing.polyNumber +
				", Color: " + nextThing.colorNumber);
	}

	boolean getPlainPiece() {
		return polyominoImage.getPlainPiece();
	}

	void setPlainPiece(boolean plainPiece) {
		polyominoImage.setPlainPiece(plainPiece);
		redrawPaint = true;
	}

	void banner(String msg) {
		((TetrisFrame) frame).displayMessage(msg);
	}

	void putUnit() {
		int i, j, xi, yj;
		int x = curThing.position.x, y = curThing.position.y;

		for (i = 0; i < curThing.size; i++)
			for (j = 0; j < curThing.size; j++) {
				xi = x + i;
				yj = y + j;
				if ((yj >= 0) &&
				    (tris[curThing.unitsIndex].poly
				     [curThing.polyNumber].shape[j][i] != 0)) {
					field[yj][xi].pmid =
					  PolyominoImage.polyPixmap(i, j, tris[curThing.unitsIndex].poly
					    [curThing.polyNumber].shape,
					    curThing.size) % PolyominoImage.NUM_BITMAPS;
					/* This uses precomputed values */
					/* field[yj][xi].pmid =
					  PolyominoImage.NUM_BITMAPS - tris[curThing.unitsIndex].poly
					   [curThing.polyNumber].shape[j][i] - 1;
					*/
					field[yj][xi].cid = curThing.colorNumber;
				}
			}
	}

	boolean overlapping() {
		int i, j, xi, yj;
		int x = curThing.position.x, y = curThing.position.y;

		for (i = 0; i < curThing.size; i++)
			for (j = 0; j < curThing.size; j++)
			  if (tris[curThing.unitsIndex].poly
			      [curThing.polyNumber].shape[j][i] != 0) {
			    xi = x + i;
			    yj = y + j;
			    if ((yj >= rows) || (xi < 0) || (xi >= columns))
			      return true;
			    if (suddenAppear) { /* creationalism */
/* This method does not allow turning polyomino to an area above screen */
			      if ((yj < 0) || (field != null && field[yj][xi].pmid >= 0))
				return true;
			    } else { /* evolution */
/* This method one can turn polyomino to an area above of screen, also
   part of the polyomino may not be visible initially */
			      if ((yj >= 0) && (field != null && field[yj][xi].pmid >= 0))
				return true;
			    }
			  }
		return false;
	}

	boolean atBottom() {
		int i, j, xi, yj;
		int x = curThing.position.x, y = curThing.position.y;

		for (i = 0; i < curThing.size; i++)
			for (j = 0; j < curThing.size; j++) {
			  xi = x + i;
			  yj = y + j;
			  if ((yj >= -1) && tris[curThing.unitsIndex].poly
			      [curThing.polyNumber].shape[j][i] != 0 &&
			      ((yj >= rows - 1) ||
			       (xi < 0) || (xi >= columns) ||
			       (field != null && field[yj + 1][xi].pmid >= 0)))
			    return true;
			}
		return false;
	}

	void tryMove(int move) {
		Thing oldThing = new Thing();
		int i;

		oldThing.clone(curThing);
		switch (move) {
		case TetrisInterface.FALL:
			curThing.position.y++;
			break;
		case TetrisInterface.DROP:
			do {
				curThing.position.y++;
				if (!dropPaused)
					score += level + fill;
			} while (!overlapping());
			curThing.position.y--;
			break;

		case TetrisInterface.MOVE_LEFT:
			curThing.position.x--;
			break;

		case TetrisInterface.MOVE_RIGHT:
			curThing.position.x++;
			break;

		case TetrisInterface.ROTATE:
			if (tris[curSpec.units - PolyominoThing.MIN_UNITS].mode
				[(curSpec.diagonal) ? 1 : 0].turnStyle > PolyominoThing.NONE) {
			  if (cw)
			    for (i = 0; i < TetrisInterface.MAX_SIDES - 1; i++)
			      curThing.polyNumber = tris[curThing.unitsIndex].poly
				[curThing.polyNumber].rotation;
			  else /* ccw */
			    curThing.polyNumber = tris[curThing.unitsIndex].poly
			      [curThing.polyNumber].rotation;
			  curThing.position.x = oldThing.position.x;
			  curThing.position.y = oldThing.position.y;
			}
			break;
		case TetrisInterface.REFLECT: /* reflect on y axis */
			if (tris[curSpec.units - PolyominoThing.MIN_UNITS].mode
			    [(curSpec.diagonal) ? 1 : 0].turnStyle == PolyominoThing.ALL) {
			  curThing.polyNumber = tris[curThing.unitsIndex].poly
			      [curThing.polyNumber].reflection;
			  curThing.position.x = oldThing.position.x;
			  curThing.position.y = oldThing.position.y;
			}
			break;
		default:
		}
		if (redrawPaint) {
			drawField();
			redrawPaint = false;
		}
		if (framePaint) {
			drawFrame(focus);
			framePaint = false;
		}
		if (overlapping()) {
			curThing = oldThing;
		} else {
			drawThingDiff(oldThing);
		}
	}

	void initFill() {
		int i, j, col, fixedUnits;
		int turnStyle = tris[curSpec.units - PolyominoThing.MIN_UNITS].mode
			[(curSpec.diagonal) ? 1 : 0].turnStyle;
		int mixed = (nextSpec.mixed && TetrisInterface.mixedLegal(nextSpec.units, nextSpec.mixed)) ? 1 : 0;
		int numberPolyominoes = tris[nextSpec.units - PolyominoThing.MIN_UNITS].mode
			[(nextSpec.diagonal) ? 1 : 0].number[mixed][turnStyle];

		if (fixedFill) {
			/* This algorithm randomly picks fixedUnits per row
			 * (4 for DEFAULTWIDTH, ie. 10/2-1)
			 * This is more in keeping with the original */
			fixedUnits = columns / 2 - 1;
			for (j = 0; j < rows; j++) {
				for (i = 0; i < columns; i++) {
					field[j][i].pmid = -1;
					field[j][i].cid = 0;
				}
				if (j >= rows - fill)
					for (i = 0; i < fixedUnits; i++) {
						do
							col = generator.nextInt(columns);
						while (field[j][col].pmid != -1);
						field[j][col].pmid = PolyominoImage.NUM_BITMAPS - 1;
						field[j][col].cid = generator.nextInt(numberPolyominoes);
					}
			}
		} else {
			/* This algorithm randomly fills boxes to a certain level
			   The problem with this is that it could potentially fill
			   up a whole row */
			for (i = 0; i < columns; i++)
				for (j = 0; j < rows; j++)
					if ((j >= rows - fill) &&
							(generator.nextInt(2) == 0)) {
						field[j][i].pmid = PolyominoImage.NUM_BITMAPS - 1;
						field[j][i].cid = generator.nextInt(numberPolyominoes);
					} else {
						field[j][i].pmid = -1;
						field[j][i].cid = 0;
					}
		}
	}

	void fillLines() {
		Graphics g = this.getGraphics();
		int i, j;
		int turnStyle = tris[curSpec.units - PolyominoThing.MIN_UNITS].mode
			[(curSpec.diagonal) ? 1 : 0].turnStyle;
		int pmid = PolyominoImage.NUM_BITMAPS - 1;

		for (j = 0; j <= rows / 2; j++) {
			for (i = 0; i < columns; i++) {
				polyominoImage.drawUnit(g, false, pmid,
				  generator.nextInt(tris[curSpec.units - PolyominoThing.MIN_UNITS].mode
				  [(curSpec.diagonal) ? 1 : 0].number[(curSpec.mixed) ? 1 : 0]
				  [turnStyle]), fieldOffset.x, fieldOffset.y,
				  i, j, boxWidth, boxHeight);
				polyominoImage.drawUnit(g, false, pmid,
				  generator.nextInt(tris[curSpec.units - PolyominoThing.MIN_UNITS].mode
				  [(curSpec.diagonal) ? 1 : 0].number[(curSpec.mixed) ? 1 : 0]
				  [turnStyle]), fieldOffset.x, fieldOffset.y,
				  i, rows - j - 1, boxWidth, boxHeight);
			}
		}
	}

	int checkLines() {
		Graphics g = this.getGraphics();
		int[] lSet = new int[rows];
		int nSet = 0, i, j, y;

		/* Find filled rows */
		for (j = 0; j < rows; j++) {
			lSet[j] = 0;
			for (i = 0; i < columns; i++)
				if (field[j][i].pmid >= 0)
					lSet[j]++;
			if (lSet[j] == columns)
				nSet++;
		}
		if (nSet == 0) {
			return 0;
		}
		for (int k = 0; k < 2 * TetrisInterface.NUM_FLASHES; k++) {
		  for (j = 0; j < rows; j++) {
		    if (lSet[j] == columns)
		      for (i = 0; i < columns; i++) {
			xorUnit(g, i, j, k);
		      }
		  }
		  try {
		    Thread.sleep(100L);
		  } catch (InterruptedException e) {
		    e.printStackTrace();
		  }
		}
		/* Shrink */
		for (j = rows - 1; j >= 0; j--)
			if (lSet[j] == columns) {
			  for (y = j; y > 0; y--)
			    for (i = 0; i < columns; i++) {
			      field[y][i].pmid = field[y - 1][i].pmid;
			      field[y][i].cid = field[y - 1][i].cid;
			    }
			  for (i = 0; i < columns; i++)
			    field[0][i].pmid = -1;
			  drawField(0, j + 1);

			  for (i = j; i > 0; i--)
			    lSet[i] = lSet[i - 1];
			  lSet[0] = 0;

			  if (j > 0)
			    for (i = 0; i < columns; i++) {
			      int tmp = field[j][i].pmid;
			      if ((tmp >= 0) && (tmp != PolyominoImage.CHECKDOWN(tmp))) {
				field[j][i].pmid = PolyominoImage.CHECKDOWN(tmp);
				polyominoImage.drawUnit(g, false,
				  field[j][i].pmid, field[j][i].cid,
				  fieldOffset.x, fieldOffset.y,
				  i, j, boxWidth, boxHeight);
			      }
			    }
			  j++;
			  if (j < rows)
			    for (i = 0; i < columns; i++) {
			      int tmp = field[j][i].pmid;

			      if ((tmp >= 0) && (tmp != PolyominoImage.CHECKUP(tmp))) {
				field[j][i].pmid = PolyominoImage.CHECKUP(tmp);
				polyominoImage.drawUnit(g, false,
				  field[j][i].pmid, field[j][i].cid,
				  fieldOffset.x, fieldOffset.y,
				  i, j, boxWidth, boxHeight);
			      }
			    }
			  if (((TetrisFrame) frame).getToggleSound())
			    ((TetrisFrame) frame).playBumpAudio();
			}
		return nSet;
	}

	@Override
	public void update(Graphics g) {
		paint(g); /* no erase */
	}

	@Override
	public void paint(Graphics g) {
		if (firstPaint) {
			firstPaint = false;
			resizePaint = true;
		}
		if (resizePaint) {
			clearField();
			resizePaint = false;
		}
		if (framePaint) {
			drawFrame(g, focus);
			framePaint = false;
		}
		redrawPaint = false;
		drawField();
		drawThing();
	}

	void paintNow() {
		Graphics g = getGraphics();

		try {
			paint(g);
		} catch (Exception e) {
			e.printStackTrace();
		} finally {
			g.dispose();
		}
	}


	transient PolyominoThing polyominoThing = null;
	transient PolyominoThingFallback polyominoThingFallback = null;
	final void readPolyominoes() {
		polyominoThing = new PolyominoThing();
		tris = PolyominoThing.readObjectsXML(TetrisInterface.XML_FILE, null);
		if (tris == null) {
			System.err.println("Using fallback pieces.");
			polyominoThingFallback = new PolyominoThingFallback();
			tris = polyominoThingFallback.readStatic();
		}
		/*if (debug) {
			checkPolyomoinoes();
		}*/
	}
}

class Spec {
	volatile int units;
	volatile boolean diagonal, mixed, bonus;

	public Spec(int units,
			boolean diagonal, boolean mixed, boolean bonus) {
		this.units = units;
		this.diagonal = diagonal;
		this.mixed = mixed;
		this.bonus = bonus;
	}

	public Spec(int units) {
		this.units = units;
		this.diagonal = this.mixed = this.bonus = false;
	}

	void clone(Spec nextSpec) {
		units = nextSpec.units;
		diagonal = nextSpec.diagonal;
		mixed = nextSpec.mixed;
		bonus = nextSpec.bonus;
	}

	public void setDiagonal(boolean diagonal) {
		this.diagonal = diagonal;
	}

	public void setMixed(boolean mixed) {
		this.mixed = mixed;
	}

	public void setBonus(boolean bonus) {
		this.bonus = bonus;
	}

	public void setUnits(int units) {
		this.units = units;
	}
}

class Field {
	int pmid, cid; // pixmaps and colors

	public Field() {
		pmid = -1;
		cid = 0;
	}
}
