package com.sillycycle.bagleyd.util;

/*
 * @(#)IODialog.java
 *
 * Structure from Java Sourcebook by Ed Anuff, pp 287-289
 *
 * Copyright 2011  David A. Bagley, bagleyd AT verizon.net
 *
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and
 * that both that copyright notice and this permission notice appear in
 * supporting documentation, and that the name of the author not be
 * used in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.
 *
 * This program is distributed in the hope that it will be "useful",
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */

import java.awt.BorderLayout;
import java.awt.Dialog;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import javax.swing.BorderFactory;
import javax.swing.JButton;
import javax.swing.JFrame;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTextArea;

/**
 * The <code>IODialog</code> class holds the I/O
 * for the <code>*Frame</code>.
 *
 * @author	David A. Bagley
 * @author	bagleyd AT verizon.net
 * @author	https://www.sillycycle.com/life.html
 */

public class IODialog extends Dialog {
	private static final long serialVersionUID = 42L;
	static boolean active = false;

	JTextArea textArea = new JTextArea("");

	public IODialog(JFrame parent,
			String titleText,
			String areaText,
			String buttonText) {
		super(parent, titleText, true);

		JButton button = new JButton(buttonText);
		JPanel buttonPanel = new JPanel();

		buttonPanel.setLayout(new ColumnLayout(0, 0, 0,
			ColumnLayout.CENTER));
		this.textArea.setFont(new Font("Courier", Font.ITALIC, 16));
		this.textArea.setLineWrap(true);
		this.textArea.setWrapStyleWord(true);
		JScrollPane areaScrollPane = new JScrollPane(this.textArea);
		/*areaScrollPane.setVerticalScrollBarPolicy(
			JScrollPane.VERTICAL_SCROLLBAR_ALWAYS);*/
		areaScrollPane.setPreferredSize(new Dimension(400, 250));
		areaScrollPane.setBorder(BorderFactory.createCompoundBorder(
			BorderFactory.createCompoundBorder(
			BorderFactory.createTitledBorder(areaText),
			BorderFactory.createEmptyBorder(5,5,5,5)),
			areaScrollPane.getBorder()));

		buttonPanel.add(areaScrollPane, BorderLayout.PAGE_START);

		buttonPanel.add(button, BorderLayout.PAGE_START);
		add(BorderLayout.PAGE_START, buttonPanel);
		pack();
		IODialog.active = true;

		button.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent event) {
				dispose();
				//else multiple Dialogs can be opened
				active = false;
			}
		});

		addWindowListener(new WindowAdapter() {
			@Override
			public void windowClosing(WindowEvent event) {
				dispose();
				//else multiple Dialogs can be opened
				active = false;
			}
		});
	}

	public void clearText() {
		this.textArea.setText("");
	}

	public void setText(String s) {
		this.textArea.setText(s);
	}

	public void appendText(String s) {
		this.textArea.append(s);
	}

	public String getText() {
		return this.textArea.getText();
	}

	public static boolean getActive() {
		return IODialog.active;
	}
}
